import os
from datetime import datetime
from email.utils import formatdate, parsedate_to_datetime

import requests
from appdirs import user_cache_dir


def download(url, destination_file):
    headers = {}

    path = user_cache_dir("promptify")

    if not os.path.isdir(path):
        os.makedirs(path)

    destination_file = os.path.join(path, destination_file)

    if os.path.exists(destination_file):
        mtime = os.path.getmtime(destination_file)
        headers["if-modified-since"] = formatdate(mtime, usegmt=True)

    response = requests.get(url, headers=headers, stream=True)
    response.raise_for_status()

    if response.status_code == requests.codes.not_modified:
        return

    if response.status_code == requests.codes.ok:
        with open(destination_file, "wb") as f:
            for chunk in response.iter_content(chunk_size=1048576):
                f.write(chunk)

        if last_modified := response.headers.get("last-modified"):
            new_mtime = parsedate_to_datetime(last_modified).timestamp()
            os.utime(destination_file, times=(datetime.now().timestamp(), new_mtime))
    return destination_file



# This file includes code which was modified from https://github.com/openai/gpt-2

import json
from functools import lru_cache

import regex as re

from .utils import download

bpe_file = {"filename": "vocab.bpe", "link": "https://github.com/syonfox/GPT-3-Encoder/raw/master/vocab.bpe"}
encoder_file = {"filename": "encoder.json", "link": "https://github.com/syonfox/GPT-3-Encoder/raw/master/encoder.json"}


@lru_cache()
def bytes_to_unicode():

    bs = (
        list(range(ord("!"), ord("~") + 1)) + list(range(ord("¡"), ord("¬") + 1)) + list(range(ord("®"), ord("ÿ") + 1))
    )
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


def get_pairs(word):
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


class Encoder:
    def __init__(self, encoder, bpe_merges, errors="replace"):
        self.encoder = encoder
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.errors = errors
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        self.bpe_ranks = dict(zip(bpe_merges, range(len(bpe_merges))))
        self.cache = {}
        self.pat = re.compile(r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""")

    def bpe(self, token):
        if token in self.cache:
            return self.cache[token]
        word = tuple(token)

        pairs = get_pairs(word)

        if not pairs:
            return token

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)

        word = " ".join(word)
        self.cache[token] = word
        return word

    def encode(self, text):
        bpe_tokens = []
        for token in re.findall(self.pat, text):
            token = "".join(self.byte_encoder[b] for b in token.encode("utf-8"))

            bpe_tokens.extend(self.encoder[bpe_token] for bpe_token in self.bpe(token).split(" "))
        return bpe_tokens

    def decode(self, tokens):
        text = "".join([self.decoder[token] for token in tokens])
        text = bytearray([self.byte_decoder[c] for c in text]).decode("utf-8", errors=self.errors)
        return text


def get_encoder():
    encoder_filename = download(encoder_file["link"], encoder_file["filename"])
    bpe_filename = download(bpe_file["link"], bpe_file["filename"])
    with open(encoder_filename, "r") as f:
        encoder = json.load(f)
    with open(bpe_filename, "r", encoding="utf-8") as f:
        bpe_data = f.read()
    bpe_merges = [tuple(merge_str.split()) for merge_str in bpe_data.split("\n")[1:-1]]
    return Encoder(encoder=encoder, bpe_merges=bpe_merges)



from typing import List


class Model:

    name = ""
    description = ""

    def __init__(
        self,
    ):
        pass

    def list_models(self) -> List[str]:
        """
        Get list of supported models
        """
        raise NotImplementedError

    def verify_model(self, model):
        """
        Verify the model is supported
        """
        return model in self.get_models()

    def run(self, *args, **kwargs):
        """
        run selected model to get completion from LLM
        """
        raise NotImplementedError



from typing import List, Union

import openai

from .model import Model
from .utils import get_encoder


class OpenAI(Model):

    name = "OpenAI"
    description = "OpenAI API for text completion using various models"

    def __init__(self, api_key: str):
        self._api_key = api_key
        self._openai = openai
        self._openai.api_key = self._api_key
        self.supported_models = ["text-davinci-003", "text-curie-001", "text-babbage-001", "text-ada-001"]
        self.encoder = get_encoder()

    def list_models(self):
        ## get all models for OpenAI API
        list_of_models = [model.id for model in self._openai.Model.list()["data"]]
        ## compare with supported models and return model_list
        models = []
        for model in self.supported_models:
            if model in list_of_models:
                models.append(model)
        return models

    def run(
        self,
        prompts: List[str],
        model_name: str = "text-davinci-003",
        temperature: float = 0.7,
        max_tokens: int = 4000,
        top_p: float = 0.1,
        frequency_penalty: float = 0,
        presence_penalty: float = 0,
        stop: Union[str, None] = None,
    ):
        """
        prompts: The prompt(s) to generate completions for, encoded as a string, array of strings, array of tokens, or array of token arrays.
        model_name: ID of the model to use. You can use the `list_models` function to see all of your available models
        temperature: What sampling temperature to use. Higher values means the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) for ones with a well-defined answer.
                    We generally recommend altering this or top_p but not both.
        max_tokens: The maximum number of tokens to generate in the completion.
                    The token count of your prompt plus max_tokens cannot exceed the model's context length. Most models have a context length of 2048 tokens (except for the newest models, which support 4096).
        top_p: An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.
                We generally recommend altering this or temperature but not both.
        frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.
        presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.
        stop: Up to 4 sequences where the API will stop generating further tokens. The returned text will not contain the stop sequence.
        """
        assert model_name in self.list_models(), "model not supported"
        result = []
        for prompt in prompts:
            len_prompt_tokens = len(self.encoder.encode(prompt))
            max_tokens_prompt = max_tokens - len_prompt_tokens
            response = self._openai.Completion.create(
                model=model_name,
                prompt=prompt,
                temperature=temperature,
                max_tokens=max_tokens_prompt,
                top_p=top_p,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                stop=stop,
            )
            data = {}
            data.update(response["usage"])
            data["text"] = response["choices"][0]["text"]
            result.append(data)

        return result



import os
from typing import List

from jinja2 import Environment, FileSystemLoader, meta

dir_path = os.path.dirname(os.path.realpath(__file__))
templates_dir = os.path.join(dir_path, "templates")


class Prompter:
    def __init__(
        self,
        model,
        templates_path=templates_dir,
        allowed_missing_variables=["examples", "description", "output_format"],
    ) -> None:
        self.environment = Environment(loader=FileSystemLoader(templates_path))
        self.model = model
        self.allowed_missing_variables = allowed_missing_variables
        self.model_args_count = self.model.run.__code__.co_argcount
        self.model_variables = self.model.run.__code__.co_varnames[1 : self.model_args_count]

    def list_templates(self) -> List[str]:
        return self.environment.list_templates()

    def get_template_variables(self, template_name: str) -> List[str]:
        template_source = self.environment.loader.get_source(self.environment, template_name)
        parsed_content = self.environment.parse(template_source)
        undeclared_variables = meta.find_undeclared_variables(parsed_content)
        return undeclared_variables

    def generate_prompt(self, template_name, **kwargs) -> str:
        variables = self.get_template_variables(template_name)
        variables_missing = []
        for variable in variables:
            if variable not in kwargs and variable not in self.allowed_missing_variables:
                variables_missing.append(variable)
        assert len(variables_missing) == 0, f"Missing required variables in template {variables_missing}"
        template = self.environment.get_template(template_name)
        prompt = template.render(**kwargs).strip()
        return prompt

    def fit(self, template_name, **kwargs):
        prompt_variables = self.get_template_variables(template_name)
        prompt_kwargs = {}
        model_kwargs = {}
        for variable in kwargs:
            if variable in prompt_variables:
                prompt_kwargs[variable] = kwargs[variable]
            elif variable in self.model_variables:
                model_kwargs[variable] = kwargs[variable]
        prompt = self.generate_prompt(template_name, **prompt_kwargs)
        output = self.model.run(prompts=[prompt], **model_kwargs)
        return output[0]
